from pytest import raises

from .. import deprecated
from ..deprecated import deprecated as deprecated_decorator
from ..deprecated import warn_deprecation


def test_warn_deprecation(mocker):
    mocker.patch.object(deprecated.warnings, "warn")

    warn_deprecation("OH!")
    deprecated.warnings.warn.assert_called_with(
        "OH!", stacklevel=2, category=DeprecationWarning
    )


def test_deprecated_decorator(mocker):
    mocker.patch.object(deprecated, "warn_deprecation")

    @deprecated_decorator
    def my_func():
        return True

    result = my_func()
    assert result
    deprecated.warn_deprecation.assert_called_with(
        "Call to deprecated function my_func."
    )


def test_deprecated_class(mocker):
    mocker.patch.object(deprecated, "warn_deprecation")

    @deprecated_decorator
    class X:
        pass

    result = X()
    assert result
    deprecated.warn_deprecation.assert_called_with("Call to deprecated class X.")


def test_deprecated_decorator_text(mocker):
    mocker.patch.object(deprecated, "warn_deprecation")

    @deprecated_decorator("Deprecation text")
    def my_func():
        return True

    result = my_func()
    assert result
    deprecated.warn_deprecation.assert_called_with(
        "Call to deprecated function my_func (Deprecation text)."
    )


def test_deprecated_class_text(mocker):
    mocker.patch.object(deprecated, "warn_deprecation")

    @deprecated_decorator("Deprecation text")
    class X:
        pass

    result = X()
    assert result
    deprecated.warn_deprecation.assert_called_with(
        "Call to deprecated class X (Deprecation text)."
    )


def test_deprecated_other_object(mocker):
    mocker.patch.object(deprecated, "warn_deprecation")

    with raises(TypeError):
        deprecated_decorator({})



from ..field import Field
from ..scalars import String


class CustomField(Field):
    def __init__(self, *args, **kwargs):
        self.metadata = kwargs.pop("metadata", None)
        super(CustomField, self).__init__(*args, **kwargs)


def test_mounted_type():
    unmounted = String()
    mounted = Field.mounted(unmounted)
    assert isinstance(mounted, Field)
    assert mounted.type == String


def test_mounted_type_custom():
    unmounted = String(metadata={"hey": "yo!"})
    mounted = CustomField.mounted(unmounted)
    assert isinstance(mounted, CustomField)
    assert mounted.type == String
    assert mounted.metadata == {"hey": "yo!"}



from ..trim_docstring import trim_docstring


def test_trim_docstring():
    class WellDocumentedObject:
        """
        This object is very well-documented. It has multiple lines in its
        description.

        Multiple paragraphs too
        """

    assert (
        trim_docstring(WellDocumentedObject.__doc__)
        == "This object is very well-documented. It has multiple lines in its\n"
        "description.\n\nMultiple paragraphs too"
    )

    class UndocumentedObject:
        pass

    assert trim_docstring(UndocumentedObject.__doc__) is None



from __future__ import absolute_import
from uuid import UUID as _UUID

from graphql.language.ast import StringValueNode
from graphql import Undefined

from .scalars import Scalar


class UUID(Scalar):
    """
    Leverages the internal Python implementation of UUID (uuid.UUID) to provide native UUID objects
    in fields, resolvers and input.
    """

    @staticmethod
    def serialize(uuid):
        if isinstance(uuid, str):
            uuid = _UUID(uuid)

        assert isinstance(uuid, _UUID), f"Expected UUID instance, received {uuid}"
        return str(uuid)

    @staticmethod
    def parse_literal(node, _variables=None):
        if isinstance(node, StringValueNode):
            return _UUID(node.value)
        return Undefined

    @staticmethod
    def parse_value(value):
        return _UUID(value)



from ..json import JSONString
from ..objecttype import ObjectType
from ..schema import Schema


class Query(ObjectType):
    json = JSONString(input=JSONString())

    def resolve_json(self, info, input):
        return input


schema = Schema(query=Query)


def test_jsonstring_query():
    json_value = '{"key": "value"}'

    json_value_quoted = json_value.replace('"', '\\"')
    result = schema.execute("""{ json(input: "%s") }""" % json_value_quoted)
    assert not result.errors
    assert result.data == {"json": json_value}

    result = schema.execute("""{ json(input: "{}") }""")
    assert not result.errors
    assert result.data == {"json": "{}"}


def test_jsonstring_query_variable():
    json_value = '{"key": "value"}'

    result = schema.execute(
        """query Test($json: JSONString){ json(input: $json) }""",
        variables={"json": json_value},
    )
    assert not result.errors
    assert result.data == {"json": json_value}


def test_jsonstring_optional_uuid_input():
    """
    Test that we can provide a null value to an optional input
    """
    result = schema.execute("{ json(input: null) }")
    assert not result.errors
    assert result.data == {"json": None}


def test_jsonstring_invalid_query():
    """
    Test that if an invalid type is provided we get an error
    """
    result = schema.execute("{ json(input: 1) }")
    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == "Expected value of type 'JSONString', found 1."

    result = schema.execute("{ json(input: {}) }")
    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == "Expected value of type 'JSONString', found {}."

    result = schema.execute('{ json(input: "a") }')
    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == (
        "Expected value of type 'JSONString', found \"a\"; "
        "Badly formed JSONString: Expecting value: line 1 column 1 (char 0)"
    )

    result = schema.execute("""{ json(input: "{\\'key\\': 0}") }""")
    assert result.errors
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "Syntax Error: Invalid character escape sequence: '\\''."
    )

    result = schema.execute("""{ json(input: "{\\"key\\": 0,}") }""")
    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == (
        'Expected value of type \'JSONString\', found "{\\"key\\": 0,}"; '
        "Badly formed JSONString: Expecting property name enclosed in double quotes: line 1 column 11 (char 10)"
    )



from __future__ import absolute_import

import datetime

from aniso8601 import parse_date, parse_datetime, parse_time
from graphql.error import GraphQLError
from graphql.language import StringValueNode, print_ast

from .scalars import Scalar


class Date(Scalar):
    """
    The `Date` scalar type represents a Date
    value as specified by
    [iso8601](https://en.wikipedia.org/wiki/ISO_8601).
    """

    @staticmethod
    def serialize(date):
        if isinstance(date, datetime.datetime):
            date = date.date()
        if not isinstance(date, datetime.date):
            raise GraphQLError(f"Date cannot represent value: {repr(date)}")
        return date.isoformat()

    @classmethod
    def parse_literal(cls, node, _variables=None):
        if not isinstance(node, StringValueNode):
            raise GraphQLError(
                f"Date cannot represent non-string value: {print_ast(node)}"
            )
        return cls.parse_value(node.value)

    @staticmethod
    def parse_value(value):
        if isinstance(value, datetime.date):
            return value
        if not isinstance(value, str):
            raise GraphQLError(f"Date cannot represent non-string value: {repr(value)}")
        try:
            return parse_date(value)
        except ValueError:
            raise GraphQLError(f"Date cannot represent value: {repr(value)}")


class DateTime(Scalar):
    """
    The `DateTime` scalar type represents a DateTime
    value as specified by
    [iso8601](https://en.wikipedia.org/wiki/ISO_8601).
    """

    @staticmethod
    def serialize(dt):
        if not isinstance(dt, (datetime.datetime, datetime.date)):
            raise GraphQLError(f"DateTime cannot represent value: {repr(dt)}")
        return dt.isoformat()

    @classmethod
    def parse_literal(cls, node, _variables=None):
        if not isinstance(node, StringValueNode):
            raise GraphQLError(
                f"DateTime cannot represent non-string value: {print_ast(node)}"
            )
        return cls.parse_value(node.value)

    @staticmethod
    def parse_value(value):
        if isinstance(value, datetime.datetime):
            return value
        if not isinstance(value, str):
            raise GraphQLError(
                f"DateTime cannot represent non-string value: {repr(value)}"
            )
        try:
            return parse_datetime(value)
        except ValueError:
            raise GraphQLError(f"DateTime cannot represent value: {repr(value)}")


class Time(Scalar):
    """
    The `Time` scalar type represents a Time value as
    specified by
    [iso8601](https://en.wikipedia.org/wiki/ISO_8601).
    """

    @staticmethod
    def serialize(time):
        if not isinstance(time, datetime.time):
            raise GraphQLError(f"Time cannot represent value: {repr(time)}")
        return time.isoformat()

    @classmethod
    def parse_literal(cls, node, _variables=None):
        if not isinstance(node, StringValueNode):
            raise GraphQLError(
                f"Time cannot represent non-string value: {print_ast(node)}"
            )
        return cls.parse_value(node.value)

    @classmethod
    def parse_value(cls, value):
        if isinstance(value, datetime.time):
            return value
        if not isinstance(value, str):
            raise GraphQLError(f"Time cannot represent non-string value: {repr(value)}")
        try:
            return parse_time(value)
        except ValueError:
            raise GraphQLError(f"Time cannot represent value: {repr(value)}")



import pytest
from graphql import Undefined

from graphene.types.inputobjecttype import set_input_object_type_default_value


@pytest.fixture()
def set_default_input_object_type_to_undefined():
    """This fixture is used to change the default value of optional inputs in InputObjectTypes for specific tests"""
    set_input_object_type_default_value(Undefined)
    yield
    set_input_object_type_default_value(None)



import decimal

from ..decimal import Decimal
from ..objecttype import ObjectType
from ..schema import Schema


class Query(ObjectType):
    decimal = Decimal(input=Decimal())

    def resolve_decimal(self, info, input):
        return input


schema = Schema(query=Query)


def test_decimal_string_query():
    decimal_value = decimal.Decimal("1969.1974")
    result = schema.execute("""{ decimal(input: "%s") }""" % decimal_value)
    assert not result.errors
    assert result.data == {"decimal": str(decimal_value)}
    assert decimal.Decimal(result.data["decimal"]) == decimal_value


def test_decimal_string_query_variable():
    decimal_value = decimal.Decimal("1969.1974")

    result = schema.execute(
        """query Test($decimal: Decimal){ decimal(input: $decimal) }""",
        variables={"decimal": decimal_value},
    )
    assert not result.errors
    assert result.data == {"decimal": str(decimal_value)}
    assert decimal.Decimal(result.data["decimal"]) == decimal_value


def test_bad_decimal_query():
    not_a_decimal = "Nobody expects the Spanish Inquisition!"

    result = schema.execute("""{ decimal(input: "%s") }""" % not_a_decimal)
    assert result.errors
    assert len(result.errors) == 1
    assert result.data is None
    assert (
        result.errors[0].message
        == "Expected value of type 'Decimal', found \"Nobody expects the Spanish Inquisition!\"."
    )

    result = schema.execute("{ decimal(input: true) }")
    assert result.errors
    assert len(result.errors) == 1
    assert result.data is None
    assert result.errors[0].message == "Expected value of type 'Decimal', found true."

    result = schema.execute("{ decimal(input: 1.2) }")
    assert result.errors
    assert len(result.errors) == 1
    assert result.data is None
    assert result.errors[0].message == "Expected value of type 'Decimal', found 1.2."


def test_decimal_string_query_integer():
    decimal_value = 1
    result = schema.execute("""{ decimal(input: %s) }""" % decimal_value)
    assert not result.errors
    assert result.data == {"decimal": str(decimal_value)}
    assert decimal.Decimal(result.data["decimal"]) == decimal_value



import re

from pytest import raises
from graphql import parse, get_introspection_query, validate

from ...types import Schema, ObjectType, Interface
from ...types import String, Int, List, Field
from ..depth_limit import depth_limit_validator


class PetType(Interface):
    name = String(required=True)

    class meta:
        name = "Pet"


class CatType(ObjectType):
    class meta:
        name = "Cat"
        interfaces = (PetType,)


class DogType(ObjectType):
    class meta:
        name = "Dog"
        interfaces = (PetType,)


class AddressType(ObjectType):
    street = String(required=True)
    number = Int(required=True)
    city = String(required=True)
    country = String(required=True)

    class Meta:
        name = "Address"


class HumanType(ObjectType):
    name = String(required=True)
    email = String(required=True)
    address = Field(AddressType, required=True)
    pets = List(PetType, required=True)

    class Meta:
        name = "Human"


class Query(ObjectType):
    user = Field(HumanType, required=True, name=String())
    version = String(required=True)
    user1 = Field(HumanType, required=True)
    user2 = Field(HumanType, required=True)
    user3 = Field(HumanType, required=True)

    @staticmethod
    def resolve_user(root, info, name=None):
        pass


schema = Schema(query=Query)


def run_query(query: str, max_depth: int, ignore=None):
    document = parse(query)

    result = None

    def callback(query_depths):
        nonlocal result
        result = query_depths

    errors = validate(
        schema=schema.graphql_schema,
        document_ast=document,
        rules=(
            depth_limit_validator(
                max_depth=max_depth, ignore=ignore, callback=callback
            ),
        ),
    )

    return errors, result


def test_should_count_depth_without_fragment():
    query = """
    query read0 {
      version
    }
    query read1 {
      version
      user {
        name
      }
    }
    query read2 {
      matt: user(name: "matt") {
        email
      }
      andy: user(name: "andy") {
        email
        address {
          city
        }
      }
    }
    query read3 {
      matt: user(name: "matt") {
        email
      }
      andy: user(name: "andy") {
        email
        address {
          city
        }
        pets {
          name
          owner {
            name
          }
        }
      }
    }
    """

    expected = {"read0": 0, "read1": 1, "read2": 2, "read3": 3}

    errors, result = run_query(query, 10)
    assert not errors
    assert result == expected


def test_should_count_with_fragments():
    query = """
    query read0 {
      ... on Query {
        version
      }
    }
    query read1 {
      version
      user {
        ... on Human {
          name
        }
      }
    }
    fragment humanInfo on Human {
      email
    }
    fragment petInfo on Pet {
      name
      owner {
        name
      }
    }
    query read2 {
      matt: user(name: "matt") {
        ...humanInfo
      }
      andy: user(name: "andy") {
        ...humanInfo
        address {
          city
        }
      }
    }
    query read3 {
      matt: user(name: "matt") {
        ...humanInfo
      }
      andy: user(name: "andy") {
        ... on Human {
          email
        }
        address {
          city
        }
        pets {
          ...petInfo
        }
      }
    }
  """

    expected = {"read0": 0, "read1": 1, "read2": 2, "read3": 3}

    errors, result = run_query(query, 10)
    assert not errors
    assert result == expected


def test_should_ignore_the_introspection_query():
    errors, result = run_query(get_introspection_query(), 10)
    assert not errors
    assert result == {"IntrospectionQuery": 0}


def test_should_catch_very_deep_query():
    query = """{
    user {
      pets {
        owner {
          pets {
            owner {
              pets {
                name
              }
            }
          }
        }
      }
    }
    }
    """
    errors, result = run_query(query, 4)

    assert len(errors) == 1
    assert errors[0].message == "'anonymous' exceeds maximum operation depth of 4."


def test_should_ignore_field():
    query = """
    query read1 {
      user { address { city } }
    }
    query read2 {
      user1 { address { city } }
      user2 { address { city } }
      user3 { address { city } }
    }
    """

    errors, result = run_query(
        query,
        10,
        ignore=["user1", re.compile("user2"), lambda field_name: field_name == "user3"],
    )

    expected = {"read1": 2, "read2": 0}
    assert not errors
    assert result == expected


def test_should_raise_invalid_ignore():
    query = """
    query read1 {
      user { address { city } }
    }
    """
    with raises(ValueError, match="Invalid ignore option:"):
        run_query(query, 10, ignore=[True])



from textwrap import dedent

from graphql import graphql_sync

from ...types import Interface, ObjectType, Schema
from ...types.scalars import Int, String
from ..node import Node


class CustomNode(Node):
    class Meta:
        name = "Node"

    @staticmethod
    def to_global_id(type_, id):
        return id

    @staticmethod
    def get_node_from_global_id(info, id, only_type=None):
        assert info.schema is graphql_schema
        if id in user_data:
            return user_data.get(id)
        else:
            return photo_data.get(id)


class BasePhoto(Interface):
    width = Int(description="The width of the photo in pixels")


class User(ObjectType):
    class Meta:
        interfaces = [CustomNode]

    name = String(description="The full name of the user")


class Photo(ObjectType):
    class Meta:
        interfaces = [CustomNode, BasePhoto]


user_data = {"1": User(id="1", name="John Doe"), "2": User(id="2", name="Jane Smith")}

photo_data = {"3": Photo(id="3", width=300), "4": Photo(id="4", width=400)}


class RootQuery(ObjectType):
    node = CustomNode.Field()


schema = Schema(query=RootQuery, types=[User, Photo])
graphql_schema = schema.graphql_schema


def test_str_schema_correct():
    assert (
        str(schema).strip()
        == dedent(
            '''
        schema {
          query: RootQuery
        }

        type User implements Node {
          """The ID of the object"""
          id: ID!

          """The full name of the user"""
          name: String
        }

        interface Node {
          """The ID of the object"""
          id: ID!
        }

        type Photo implements Node & BasePhoto {
          """The ID of the object"""
          id: ID!

          """The width of the photo in pixels"""
          width: Int
        }

        interface BasePhoto {
          """The width of the photo in pixels"""
          width: Int
        }

        type RootQuery {
          node(
            """The ID of the object"""
            id: ID!
          ): Node
        }
        '''
        ).strip()
    )


def test_gets_the_correct_id_for_users():
    query = """
      {
        node(id: "1") {
          id
        }
      }
    """
    expected = {"node": {"id": "1"}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_gets_the_correct_id_for_photos():
    query = """
      {
        node(id: "4") {
          id
        }
      }
    """
    expected = {"node": {"id": "4"}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_gets_the_correct_name_for_users():
    query = """
      {
        node(id: "1") {
          id
          ... on User {
            name
          }
        }
      }
    """
    expected = {"node": {"id": "1", "name": "John Doe"}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_gets_the_correct_width_for_photos():
    query = """
      {
        node(id: "4") {
          id
          ... on Photo {
            width
          }
        }
      }
    """
    expected = {"node": {"id": "4", "width": 400}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_gets_the_correct_typename_for_users():
    query = """
      {
        node(id: "1") {
          id
          __typename
        }
      }
    """
    expected = {"node": {"id": "1", "__typename": "User"}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_gets_the_correct_typename_for_photos():
    query = """
      {
        node(id: "4") {
          id
          __typename
        }
      }
    """
    expected = {"node": {"id": "4", "__typename": "Photo"}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_ignores_photo_fragments_on_user():
    query = """
      {
        node(id: "1") {
          id
          ... on Photo {
            width
          }
        }
      }
    """
    expected = {"node": {"id": "1"}}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_returns_null_for_bad_ids():
    query = """
      {
        node(id: "5") {
          id
        }
      }
    """
    expected = {"node": None}
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_have_correct_node_interface():
    query = """
      {
        __type(name: "Node") {
          name
          kind
          fields {
            name
            type {
              kind
              ofType {
                name
                kind
              }
            }
          }
        }
      }
    """
    expected = {
        "__type": {
            "name": "Node",
            "kind": "INTERFACE",
            "fields": [
                {
                    "name": "id",
                    "type": {
                        "kind": "NON_NULL",
                        "ofType": {"name": "ID", "kind": "SCALAR"},
                    },
                }
            ],
        }
    }
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected


def test_has_correct_node_root_field():
    query = """
      {
        __schema {
          queryType {
            fields {
              name
              type {
                name
                kind
              }
              args {
                name
                type {
                  kind
                  ofType {
                    name
                    kind
                  }
                }
              }
            }
          }
        }
      }
    """
    expected = {
        "__schema": {
            "queryType": {
                "fields": [
                    {
                        "name": "node",
                        "type": {"name": "Node", "kind": "INTERFACE"},
                        "args": [
                            {
                                "name": "id",
                                "type": {
                                    "kind": "NON_NULL",
                                    "ofType": {"name": "ID", "kind": "SCALAR"},
                                },
                            }
                        ],
                    }
                ]
            }
        }
    }
    result = graphql_sync(graphql_schema, query)
    assert not result.errors
    assert result.data == expected



from functools import partial

from pytest import raises

from ..inputfield import InputField
from ..structures import NonNull
from .utils import MyLazyType


def test_inputfield_required():
    MyType = object()
    field = InputField(MyType, required=True)
    assert isinstance(field.type, NonNull)
    assert field.type.of_type == MyType


def test_inputfield_deprecated():
    MyType = object()
    deprecation_reason = "deprecated"
    field = InputField(MyType, required=False, deprecation_reason=deprecation_reason)
    assert isinstance(field.type, type(MyType))
    assert field.deprecation_reason == deprecation_reason


def test_inputfield_required_deprecated():
    MyType = object()
    with raises(AssertionError) as exc_info:
        InputField(MyType, name="input", required=True, deprecation_reason="deprecated")

    assert str(exc_info.value) == "InputField input is required, cannot deprecate it."


def test_inputfield_with_lazy_type():
    MyType = object()
    field = InputField(lambda: MyType)
    assert field.type == MyType


def test_inputfield_with_lazy_partial_type():
    MyType = object()
    field = InputField(partial(lambda: MyType))
    assert field.type == MyType


def test_inputfield_with_string_type():
    field = InputField("graphene.types.tests.utils.MyLazyType")
    assert field.type == MyLazyType



from pytest import mark

from graphql_relay.utils import base64

from graphene.types import ObjectType, Schema, String
from graphene.relay.connection import Connection, ConnectionField, PageInfo
from graphene.relay.node import Node

letter_chars = ["A", "B", "C", "D", "E"]


class Letter(ObjectType):
    class Meta:
        interfaces = (Node,)

    letter = String()


class LetterConnection(Connection):
    class Meta:
        node = Letter


class Query(ObjectType):
    letters = ConnectionField(LetterConnection)
    connection_letters = ConnectionField(LetterConnection)
    async_letters = ConnectionField(LetterConnection)

    node = Node.Field()

    def resolve_letters(self, info, **args):
        return list(letters.values())

    async def resolve_async_letters(self, info, **args):
        return list(letters.values())

    def resolve_connection_letters(self, info, **args):
        return LetterConnection(
            page_info=PageInfo(has_next_page=True, has_previous_page=False),
            edges=[
                LetterConnection.Edge(node=Letter(id=0, letter="A"), cursor="a-cursor")
            ],
        )


schema = Schema(Query)

letters = {letter: Letter(id=i, letter=letter) for i, letter in enumerate(letter_chars)}


def edges(selected_letters):
    return [
        {
            "node": {"id": base64("Letter:%s" % letter.id), "letter": letter.letter},
            "cursor": base64("arrayconnection:%s" % letter.id),
        }
        for letter in [letters[i] for i in selected_letters]
    ]


def cursor_for(ltr):
    letter = letters[ltr]
    return base64("arrayconnection:%s" % letter.id)


def execute(args=""):
    if args:
        args = "(" + args + ")"

    return schema.execute(
        """
    {
        letters%s {
            edges {
                node {
                    id
                    letter
                }
                cursor
            }
            pageInfo {
                hasPreviousPage
                hasNextPage
                startCursor
                endCursor
            }
        }
    }
    """
        % args
    )


@mark.asyncio
async def test_connection_async():
    result = await schema.execute_async(
        """
    {
        asyncLetters(first:1) {
            edges {
                node {
                    id
                    letter
                }
            }
            pageInfo {
                hasPreviousPage
                hasNextPage
            }
        }
    }
    """
    )

    assert not result.errors
    assert result.data == {
        "asyncLetters": {
            "edges": [{"node": {"id": "TGV0dGVyOjA=", "letter": "A"}}],
            "pageInfo": {"hasPreviousPage": False, "hasNextPage": True},
        }
    }



from typing import TYPE_CHECKING

from .base import BaseOptions, BaseType
from .unmountedtype import UnmountedType

# For static type checking with type checker
if TYPE_CHECKING:
    from .objecttype import ObjectType  # NOQA
    from typing import Iterable, Type  # NOQA


class UnionOptions(BaseOptions):
    types = ()  # type: Iterable[Type[ObjectType]]


class Union(UnmountedType, BaseType):
    """
    Union Type Definition

    When a field can return one of a heterogeneous set of types, a Union type
    is used to describe what types are possible as well as providing a function
    to determine which type is actually used when the field is resolved.

    The schema in this example can take a search text and return any of the GraphQL object types
    indicated: Human, Droid or Starship.

    Ambiguous return types can be resolved on each ObjectType through ``Meta.possible_types``
    attribute or ``is_type_of`` method. Or by implementing ``resolve_type`` class method on the
    Union.

    .. code:: python

        from graphene import Union, ObjectType, List

        class SearchResult(Union):
            class Meta:
                types = (Human, Droid, Starship)

        class Query(ObjectType):
            search = List(SearchResult.Field(
                search_text=String(description='Value to search for'))
            )

    Meta:
        types (Iterable[graphene.ObjectType]): Required. Collection of types that may be returned
            by this Union for the graphQL schema.
        name (optional, str): the name of the GraphQL type (must be unique in schema). Defaults to class
            name.
        description (optional, str): the description of the GraphQL type in the schema. Defaults to class
            docstring.
    """

    @classmethod
    def __init_subclass_with_meta__(cls, types=None, **options):
        assert (
            isinstance(types, (list, tuple)) and len(types) > 0
        ), f"Must provide types for Union {cls.__name__}."

        _meta = UnionOptions(cls)
        _meta.types = types
        super(Union, cls).__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    def get_type(cls):
        """
        This function is called when the unmounted type (Union instance)
        is mounted (as a Field, InputField or Argument)
        """
        return cls

    @classmethod
    def resolve_type(cls, instance, info):
        from .objecttype import ObjectType  # NOQA

        if isinstance(instance, ObjectType):
            return type(instance)



from functools import partial

from pytest import raises

from ..argument import Argument, to_arguments
from ..field import Field
from ..inputfield import InputField
from ..scalars import String
from ..structures import NonNull


def test_argument():
    arg = Argument(String, default_value="a", description="desc", name="b")
    assert arg.type == String
    assert arg.default_value == "a"
    assert arg.description == "desc"
    assert arg.name == "b"


def test_argument_comparasion():
    arg1 = Argument(
        String,
        name="Hey",
        description="Desc",
        default_value="default",
        deprecation_reason="deprecated",
    )
    arg2 = Argument(
        String,
        name="Hey",
        description="Desc",
        default_value="default",
        deprecation_reason="deprecated",
    )

    assert arg1 == arg2
    assert arg1 != String()


def test_argument_required():
    arg = Argument(String, required=True)
    assert arg.type == NonNull(String)


def test_to_arguments():
    args = {"arg_string": Argument(String), "unmounted_arg": String(required=True)}

    my_args = to_arguments(args)
    assert my_args == {
        "arg_string": Argument(String),
        "unmounted_arg": Argument(String, required=True),
    }


def test_to_arguments_deprecated():
    args = {"unmounted_arg": String(required=False, deprecation_reason="deprecated")}

    my_args = to_arguments(args)
    assert my_args == {
        "unmounted_arg": Argument(
            String, required=False, deprecation_reason="deprecated"
        ),
    }


def test_to_arguments_required_deprecated():
    args = {
        "unmounted_arg": String(
            required=True, name="arg", deprecation_reason="deprecated"
        )
    }

    with raises(AssertionError) as exc_info:
        to_arguments(args)

    assert str(exc_info.value) == "Argument arg is required, cannot deprecate it."


def test_to_arguments_raises_if_field():
    args = {"arg_string": Field(String)}

    with raises(ValueError) as exc_info:
        to_arguments(args)

    assert str(exc_info.value) == (
        "Expected arg_string to be Argument, but received Field. Try using "
        "Argument(String)."
    )


def test_to_arguments_raises_if_inputfield():
    args = {"arg_string": InputField(String)}

    with raises(ValueError) as exc_info:
        to_arguments(args)

    assert str(exc_info.value) == (
        "Expected arg_string to be Argument, but received InputField. Try "
        "using Argument(String)."
    )


def test_argument_with_lazy_type():
    MyType = object()
    arg = Argument(lambda: MyType)
    assert arg.type == MyType


def test_argument_with_lazy_partial_type():
    MyType = object()
    arg = Argument(partial(lambda: MyType))
    assert arg.type == MyType



from ..objecttype import ObjectType
from ..schema import Schema
from ..uuid import UUID
from ..structures import NonNull


class Query(ObjectType):
    uuid = UUID(input=UUID())
    required_uuid = UUID(input=NonNull(UUID), required=True)

    def resolve_uuid(self, info, input):
        return input

    def resolve_required_uuid(self, info, input):
        return input


schema = Schema(query=Query)


def test_uuidstring_query():
    uuid_value = "dfeb3bcf-70fd-11e7-a61a-6003088f8204"
    result = schema.execute("""{ uuid(input: "%s") }""" % uuid_value)
    assert not result.errors
    assert result.data == {"uuid": uuid_value}


def test_uuidstring_query_variable():
    uuid_value = "dfeb3bcf-70fd-11e7-a61a-6003088f8204"

    result = schema.execute(
        """query Test($uuid: UUID){ uuid(input: $uuid) }""",
        variables={"uuid": uuid_value},
    )
    assert not result.errors
    assert result.data == {"uuid": uuid_value}


def test_uuidstring_optional_uuid_input():
    """
    Test that we can provide a null value to an optional input
    """
    result = schema.execute("{ uuid(input: null) }")
    assert not result.errors
    assert result.data == {"uuid": None}


def test_uuidstring_invalid_query():
    """
    Test that if an invalid type is provided we get an error
    """
    result = schema.execute("{ uuid(input: 1) }")
    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == "Expected value of type 'UUID', found 1."

    result = schema.execute('{ uuid(input: "a") }')
    assert result.errors
    assert len(result.errors) == 1
    assert (
        result.errors[0].message
        == "Expected value of type 'UUID', found \"a\"; badly formed hexadecimal UUID string"
    )

    result = schema.execute("{ requiredUuid(input: null) }")
    assert result.errors
    assert len(result.errors) == 1
    assert result.errors[0].message == "Expected value of type 'UUID!', found null."



from graphql_relay import to_global_id

from ...types import ID, NonNull, ObjectType, String
from ...types.definitions import GrapheneObjectType
from ..node import GlobalID, Node


class CustomNode(Node):
    class Meta:
        name = "Node"


class User(ObjectType):
    class Meta:
        interfaces = [CustomNode]

    name = String()


class Info:
    def __init__(self, parent_type):
        self.parent_type = GrapheneObjectType(
            graphene_type=parent_type,
            name=parent_type._meta.name,
            description=parent_type._meta.description,
            fields=None,
            is_type_of=parent_type.is_type_of,
            interfaces=None,
        )


def test_global_id_defaults_to_required_and_node():
    gid = GlobalID()
    assert isinstance(gid.type, NonNull)
    assert gid.type.of_type == ID
    assert gid.node == Node


def test_global_id_allows_overriding_of_node_and_required():
    gid = GlobalID(node=CustomNode, required=False)
    assert gid.type == ID
    assert gid.node == CustomNode


def test_global_id_defaults_to_info_parent_type():
    my_id = "1"
    gid = GlobalID()
    id_resolver = gid.wrap_resolve(lambda *_: my_id)
    my_global_id = id_resolver(None, Info(User))
    assert my_global_id == to_global_id(User._meta.name, my_id)


def test_global_id_allows_setting_customer_parent_type():
    my_id = "1"
    gid = GlobalID(parent_type=User)
    id_resolver = gid.wrap_resolve(lambda *_: my_id)
    my_global_id = id_resolver(None, None)
    assert my_global_id == to_global_id(User._meta.name, my_id)



import copy

from ..argument import Argument
from ..definitions import GrapheneGraphQLType
from ..enum import Enum
from ..field import Field
from ..inputfield import InputField
from ..inputobjecttype import InputObjectType
from ..interface import Interface
from ..objecttype import ObjectType
from ..scalars import Boolean, Int, String
from ..schema import Schema
from ..structures import List, NonNull
from ..union import Union


class Image(ObjectType):
    url = String()
    width = Int()
    height = Int()


class Author(ObjectType):
    id = String()
    name = String()
    pic = Field(Image, width=Int(), height=Int())
    recent_article = Field(lambda: Article)


class Article(ObjectType):
    id = String()
    is_published = Boolean()
    author = Field(Author)
    title = String()
    body = String()


class Query(ObjectType):
    article = Field(Article, id=String())
    feed = List(Article)


class Mutation(ObjectType):
    write_article = Field(Article)


class Subscription(ObjectType):
    article_subscribe = Field(Article, id=String())


class MyObjectType(ObjectType):
    pass


class MyInterface(Interface):
    pass


class MyUnion(Union):
    class Meta:
        types = (Article,)


class MyEnum(Enum):
    foo = "foo"


class MyInputObjectType(InputObjectType):
    pass


def test_defines_a_query_only_schema():
    blog_schema = Schema(Query)

    assert blog_schema.query == Query
    assert blog_schema.graphql_schema.query_type.graphene_type == Query

    article_field = Query._meta.fields["article"]
    assert article_field.type == Article
    assert article_field.type._meta.name == "Article"

    article_field_type = article_field.type
    assert issubclass(article_field_type, ObjectType)

    title_field = article_field_type._meta.fields["title"]
    assert title_field.type == String

    author_field = article_field_type._meta.fields["author"]
    author_field_type = author_field.type
    assert issubclass(author_field_type, ObjectType)
    recent_article_field = author_field_type._meta.fields["recent_article"]

    assert recent_article_field.type == Article

    feed_field = Query._meta.fields["feed"]
    assert feed_field.type.of_type == Article


def test_defines_a_mutation_schema():
    blog_schema = Schema(Query, mutation=Mutation)

    assert blog_schema.mutation == Mutation
    assert blog_schema.graphql_schema.mutation_type.graphene_type == Mutation

    write_mutation = Mutation._meta.fields["write_article"]
    assert write_mutation.type == Article
    assert write_mutation.type._meta.name == "Article"


def test_defines_a_subscription_schema():
    blog_schema = Schema(Query, subscription=Subscription)

    assert blog_schema.subscription == Subscription
    assert blog_schema.graphql_schema.subscription_type.graphene_type == Subscription

    subscription = Subscription._meta.fields["article_subscribe"]
    assert subscription.type == Article
    assert subscription.type._meta.name == "Article"


def test_includes_nested_input_objects_in_the_map():
    class NestedInputObject(InputObjectType):
        value = String()

    class SomeInputObject(InputObjectType):
        nested = InputField(NestedInputObject)

    class SomeMutation(Mutation):
        mutate_something = Field(Article, input=Argument(SomeInputObject))

    class SomeSubscription(Mutation):
        subscribe_to_something = Field(Article, input=Argument(SomeInputObject))

    schema = Schema(query=Query, mutation=SomeMutation, subscription=SomeSubscription)
    type_map = schema.graphql_schema.type_map

    assert type_map["NestedInputObject"].graphene_type is NestedInputObject


def test_includes_interfaces_thunk_subtypes_in_the_type_map():
    class SomeInterface(Interface):
        f = Int()

    class SomeSubtype(ObjectType):
        class Meta:
            interfaces = (SomeInterface,)

    class Query(ObjectType):
        iface = Field(lambda: SomeInterface)

    schema = Schema(query=Query, types=[SomeSubtype])
    type_map = schema.graphql_schema.type_map

    assert type_map["SomeSubtype"].graphene_type is SomeSubtype


def test_includes_types_in_union():
    class SomeType(ObjectType):
        a = String()

    class OtherType(ObjectType):
        b = String()

    class MyUnion(Union):
        class Meta:
            types = (SomeType, OtherType)

    class Query(ObjectType):
        union = Field(MyUnion)

    schema = Schema(query=Query)
    type_map = schema.graphql_schema.type_map

    assert type_map["OtherType"].graphene_type is OtherType
    assert type_map["SomeType"].graphene_type is SomeType


def test_maps_enum():
    class SomeType(ObjectType):
        a = String()

    class OtherType(ObjectType):
        b = String()

    class MyUnion(Union):
        class Meta:
            types = (SomeType, OtherType)

    class Query(ObjectType):
        union = Field(MyUnion)

    schema = Schema(query=Query)
    type_map = schema.graphql_schema.type_map

    assert type_map["OtherType"].graphene_type is OtherType
    assert type_map["SomeType"].graphene_type is SomeType


def test_includes_interfaces_subtypes_in_the_type_map():
    class SomeInterface(Interface):
        f = Int()

    class SomeSubtype(ObjectType):
        class Meta:
            interfaces = (SomeInterface,)

    class Query(ObjectType):
        iface = Field(SomeInterface)

    schema = Schema(query=Query, types=[SomeSubtype])
    type_map = schema.graphql_schema.type_map

    assert type_map["SomeSubtype"].graphene_type is SomeSubtype


def test_stringifies_simple_types():
    assert str(Int) == "Int"
    assert str(Article) == "Article"
    assert str(MyInterface) == "MyInterface"
    assert str(MyUnion) == "MyUnion"
    assert str(MyEnum) == "MyEnum"
    assert str(MyInputObjectType) == "MyInputObjectType"
    assert str(NonNull(Int)) == "Int!"
    assert str(List(Int)) == "[Int]"
    assert str(NonNull(List(Int))) == "[Int]!"
    assert str(List(NonNull(Int))) == "[Int!]"
    assert str(List(List(Int))) == "[[Int]]"


# def test_identifies_input_types():
#     expected = (
#         (GraphQLInt, True),
#         (ObjectType, False),
#         (InterfaceType, False),
#         (UnionType, False),
#         (EnumType, True),
#         (InputObjectType, True)
#     )

#     for type_, answer in expected:
#         assert is_input_type(type_) == answer
#         assert is_input_type(GraphQLList(type_)) == answer
#         assert is_input_type(GraphQLNonNull(type_)) == answer


# def test_identifies_output_types():
#     expected = (
#         (GraphQLInt, True),
#         (ObjectType, True),
#         (InterfaceType, True),
#         (UnionType, True),
#         (EnumType, True),
#         (InputObjectType, False)
#     )

#     for type, answer in expected:
#         assert is_output_type(type) == answer
#         assert is_output_type(GraphQLList(type)) == answer
#         assert is_output_type(GraphQLNonNull(type)) == answer


# def test_prohibits_nesting_nonnull_inside_nonnull():
#     with raises(Exception) as excinfo:
#         GraphQLNonNull(GraphQLNonNull(GraphQLInt))

#     assert 'Can only create NonNull of a Nullable GraphQLType but got: Int!.' in str(excinfo.value)


# def test_prohibits_putting_non_object_types_in_unions():
#     bad_union_types = [
#         GraphQLInt,
#         GraphQLNonNull(GraphQLInt),
#         GraphQLList(GraphQLInt),
#         InterfaceType,
#         UnionType,
#         EnumType,
#         InputObjectType
#     ]
#     for x in bad_union_types:
#         with raises(Exception) as excinfo:
#             GraphQLSchema(
#                 GraphQLObjectType(
#                     'Root',
#                     fields={
#                         'union': GraphQLField(GraphQLUnionType('BadUnion', [x]))
#                     }
#                 )
#             )

#         assert 'BadUnion may only contain Object types, it cannot contain: ' + str(x) + '.' \
#                == str(excinfo.value)


def test_does_not_mutate_passed_field_definitions():
    class CommonFields:
        field1 = String()
        field2 = String(id=String())

    class TestObject1(CommonFields, ObjectType):
        pass

    class TestObject2(CommonFields, ObjectType):
        pass

    assert TestObject1._meta.fields == TestObject2._meta.fields

    class CommonFields:
        field1 = String()
        field2 = String()

    class TestInputObject1(CommonFields, InputObjectType):
        pass

    class TestInputObject2(CommonFields, InputObjectType):
        pass

    assert TestInputObject1._meta.fields == TestInputObject2._meta.fields


def test_graphene_graphql_type_can_be_copied():
    class Query(ObjectType):
        field = String()

        def resolve_field(self, info):
            return ""

    schema = Schema(query=Query)
    query_type_copy = copy.copy(schema.graphql_schema.query_type)
    assert query_type_copy.__dict__ == schema.graphql_schema.query_type.__dict__
    assert isinstance(schema.graphql_schema.query_type, GrapheneGraphQLType)



from graphql import parse, validate

from ...types import Schema, ObjectType, String
from ..disable_introspection import DisableIntrospection


class Query(ObjectType):
    name = String(required=True)

    @staticmethod
    def resolve_name(root, info):
        return "Hello world!"


schema = Schema(query=Query)


def run_query(query: str):
    document = parse(query)

    return validate(
        schema=schema.graphql_schema,
        document_ast=document,
        rules=(DisableIntrospection,),
    )


def test_disallows_introspection_queries():
    errors = run_query("{ __schema { queryType { name } } }")

    assert len(errors) == 1
    assert errors[0].message == "Cannot query '__schema': introspection is disabled."


def test_allows_non_introspection_queries():
    errors = run_query("{ name }")
    assert len(errors) == 0



from pytest import mark, raises

from ...types import (
    ID,
    Argument,
    Field,
    InputField,
    InputObjectType,
    NonNull,
    ObjectType,
    Schema,
)
from ...types.scalars import String
from ..mutation import ClientIDMutation


class SharedFields:
    shared = String()


class MyNode(ObjectType):
    # class Meta:
    #     interfaces = (Node, )
    id = ID()
    name = String()


class SaySomething(ClientIDMutation):
    class Input:
        what = String()

    phrase = String()

    @staticmethod
    def mutate_and_get_payload(self, info, what, client_mutation_id=None):
        return SaySomething(phrase=str(what))


class FixedSaySomething:
    __slots__ = ("phrase",)

    def __init__(self, phrase):
        self.phrase = phrase


class SaySomethingFixed(ClientIDMutation):
    class Input:
        what = String()

    phrase = String()

    @staticmethod
    def mutate_and_get_payload(self, info, what, client_mutation_id=None):
        return FixedSaySomething(phrase=str(what))


class SaySomethingAsync(ClientIDMutation):
    class Input:
        what = String()

    phrase = String()

    @staticmethod
    async def mutate_and_get_payload(self, info, what, client_mutation_id=None):
        return SaySomething(phrase=str(what))


# MyEdge = MyNode.Connection.Edge
class MyEdge(ObjectType):
    node = Field(MyNode)
    cursor = String()


class OtherMutation(ClientIDMutation):
    class Input(SharedFields):
        additional_field = String()

    name = String()
    my_node_edge = Field(MyEdge)

    @staticmethod
    def mutate_and_get_payload(
        self, info, shared="", additional_field="", client_mutation_id=None
    ):
        edge_type = MyEdge
        return OtherMutation(
            name=shared + additional_field,
            my_node_edge=edge_type(cursor="1", node=MyNode(name="name")),
        )


class RootQuery(ObjectType):
    something = String()


class Mutation(ObjectType):
    say = SaySomething.Field()
    say_fixed = SaySomethingFixed.Field()
    say_async = SaySomethingAsync.Field()
    other = OtherMutation.Field()


schema = Schema(query=RootQuery, mutation=Mutation)


def test_no_mutate_and_get_payload():
    with raises(AssertionError) as excinfo:

        class MyMutation(ClientIDMutation):
            pass

    assert (
        "MyMutation.mutate_and_get_payload method is required in a ClientIDMutation."
        == str(excinfo.value)
    )


def test_mutation():
    fields = SaySomething._meta.fields
    assert list(fields) == ["phrase", "client_mutation_id"]
    assert SaySomething._meta.name == "SaySomethingPayload"
    assert isinstance(fields["phrase"], Field)
    field = SaySomething.Field()
    assert field.type == SaySomething
    assert list(field.args) == ["input"]
    assert isinstance(field.args["input"], Argument)
    assert isinstance(field.args["input"].type, NonNull)
    assert field.args["input"].type.of_type == SaySomething.Input
    assert isinstance(fields["client_mutation_id"], Field)
    assert fields["client_mutation_id"].name == "clientMutationId"
    assert fields["client_mutation_id"].type == String


def test_mutation_input():
    Input = SaySomething.Input
    assert issubclass(Input, InputObjectType)
    fields = Input._meta.fields
    assert list(fields) == ["what", "client_mutation_id"]
    assert isinstance(fields["what"], InputField)
    assert fields["what"].type == String
    assert isinstance(fields["client_mutation_id"], InputField)
    assert fields["client_mutation_id"].type == String


def test_subclassed_mutation():
    fields = OtherMutation._meta.fields
    assert list(fields) == ["name", "my_node_edge", "client_mutation_id"]
    assert isinstance(fields["name"], Field)
    field = OtherMutation.Field()
    assert field.type == OtherMutation
    assert list(field.args) == ["input"]
    assert isinstance(field.args["input"], Argument)
    assert isinstance(field.args["input"].type, NonNull)
    assert field.args["input"].type.of_type == OtherMutation.Input


def test_subclassed_mutation_input():
    Input = OtherMutation.Input
    assert issubclass(Input, InputObjectType)
    fields = Input._meta.fields
    assert list(fields) == ["shared", "additional_field", "client_mutation_id"]
    assert isinstance(fields["shared"], InputField)
    assert fields["shared"].type == String
    assert isinstance(fields["additional_field"], InputField)
    assert fields["additional_field"].type == String
    assert isinstance(fields["client_mutation_id"], InputField)
    assert fields["client_mutation_id"].type == String


def test_node_query():
    executed = schema.execute(
        'mutation a { say(input: {what:"hello", clientMutationId:"1"}) { phrase } }'
    )
    assert not executed.errors
    assert executed.data == {"say": {"phrase": "hello"}}


def test_node_query_fixed():
    executed = schema.execute(
        'mutation a { sayFixed(input: {what:"hello", clientMutationId:"1"}) { phrase } }'
    )
    assert "Cannot set client_mutation_id in the payload object" in str(
        executed.errors[0]
    )


@mark.asyncio
async def test_node_query_async():
    executed = await schema.execute_async(
        'mutation a { sayAsync(input: {what:"hello", clientMutationId:"1"}) { phrase } }'
    )
    assert not executed.errors
    assert executed.data == {"sayAsync": {"phrase": "hello"}}


def test_edge_query():
    executed = schema.execute(
        'mutation a { other(input: {clientMutationId:"1"}) { clientMutationId, myNodeEdge { cursor node { name }} } }'
    )
    assert not executed.errors
    assert dict(executed.data) == {
        "other": {
            "clientMutationId": "1",
            "myNodeEdge": {"cursor": "1", "node": {"name": "name"}},
        }
    }



from pytest import mark

from graphene.types import ID, Field, ObjectType, Schema
from graphene.types.scalars import String
from graphene.relay.mutation import ClientIDMutation
from graphene.test import Client


class SharedFields(object):
    shared = String()


class MyNode(ObjectType):
    # class Meta:
    #     interfaces = (Node, )
    id = ID()
    name = String()


class SaySomethingAsync(ClientIDMutation):
    class Input:
        what = String()

    phrase = String()

    @staticmethod
    async def mutate_and_get_payload(self, info, what, client_mutation_id=None):
        return SaySomethingAsync(phrase=str(what))


# MyEdge = MyNode.Connection.Edge
class MyEdge(ObjectType):
    node = Field(MyNode)
    cursor = String()


class OtherMutation(ClientIDMutation):
    class Input(SharedFields):
        additional_field = String()

    name = String()
    my_node_edge = Field(MyEdge)

    @staticmethod
    def mutate_and_get_payload(
        self, info, shared="", additional_field="", client_mutation_id=None
    ):
        edge_type = MyEdge
        return OtherMutation(
            name=shared + additional_field,
            my_node_edge=edge_type(cursor="1", node=MyNode(name="name")),
        )


class RootQuery(ObjectType):
    something = String()


class Mutation(ObjectType):
    say_promise = SaySomethingAsync.Field()
    other = OtherMutation.Field()


schema = Schema(query=RootQuery, mutation=Mutation)
client = Client(schema)


@mark.asyncio
async def test_node_query_promise():
    executed = await client.execute_async(
        'mutation a { sayPromise(input: {what:"hello", clientMutationId:"1"}) { phrase } }'
    )
    assert isinstance(executed, dict)
    assert "errors" not in executed
    assert executed["data"] == {"sayPromise": {"phrase": "hello"}}


@mark.asyncio
async def test_edge_query():
    executed = await client.execute_async(
        'mutation a { other(input: {clientMutationId:"1"}) { clientMutationId, myNodeEdge { cursor node { name }} } }'
    )
    assert isinstance(executed, dict)
    assert "errors" not in executed
    assert executed["data"] == {
        "other": {
            "clientMutationId": "1",
            "myNodeEdge": {"cursor": "1", "node": {"name": "name"}},
        }
    }



from graphql import Undefined

from ..argument import Argument
from ..field import Field
from ..inputfield import InputField
from ..inputobjecttype import InputObjectType
from ..objecttype import ObjectType
from ..scalars import Boolean, String
from ..schema import Schema
from ..unmountedtype import UnmountedType
from ... import NonNull


class MyType:
    pass


class MyScalar(UnmountedType):
    def get_type(self):
        return MyType


def test_generate_inputobjecttype():
    class MyInputObjectType(InputObjectType):
        """Documentation"""

    assert MyInputObjectType._meta.name == "MyInputObjectType"
    assert MyInputObjectType._meta.description == "Documentation"
    assert MyInputObjectType._meta.fields == {}


def test_generate_inputobjecttype_with_meta():
    class MyInputObjectType(InputObjectType):
        class Meta:
            name = "MyOtherInputObjectType"
            description = "Documentation"

    assert MyInputObjectType._meta.name == "MyOtherInputObjectType"
    assert MyInputObjectType._meta.description == "Documentation"


def test_generate_inputobjecttype_with_fields():
    class MyInputObjectType(InputObjectType):
        field = Field(MyType)

    assert "field" in MyInputObjectType._meta.fields


def test_ordered_fields_in_inputobjecttype():
    class MyInputObjectType(InputObjectType):
        b = InputField(MyType)
        a = InputField(MyType)
        field = MyScalar()
        asa = InputField(MyType)

    assert list(MyInputObjectType._meta.fields) == ["b", "a", "field", "asa"]


def test_generate_inputobjecttype_unmountedtype():
    class MyInputObjectType(InputObjectType):
        field = MyScalar(MyType)

    assert "field" in MyInputObjectType._meta.fields
    assert isinstance(MyInputObjectType._meta.fields["field"], InputField)


def test_generate_inputobjecttype_as_argument():
    class MyInputObjectType(InputObjectType):
        field = MyScalar()

    class MyObjectType(ObjectType):
        field = Field(MyType, input=MyInputObjectType())

    assert "field" in MyObjectType._meta.fields
    field = MyObjectType._meta.fields["field"]
    assert isinstance(field, Field)
    assert field.type == MyType
    assert "input" in field.args
    assert isinstance(field.args["input"], Argument)
    assert field.args["input"].type == MyInputObjectType


def test_generate_inputobjecttype_inherit_abstracttype():
    class MyAbstractType:
        field1 = MyScalar(MyType)

    class MyInputObjectType(InputObjectType, MyAbstractType):
        field2 = MyScalar(MyType)

    assert list(MyInputObjectType._meta.fields) == ["field1", "field2"]
    assert [type(x) for x in MyInputObjectType._meta.fields.values()] == [
        InputField,
        InputField,
    ]


def test_generate_inputobjecttype_inherit_abstracttype_reversed():
    class MyAbstractType:
        field1 = MyScalar(MyType)

    class MyInputObjectType(MyAbstractType, InputObjectType):
        field2 = MyScalar(MyType)

    assert list(MyInputObjectType._meta.fields) == ["field1", "field2"]
    assert [type(x) for x in MyInputObjectType._meta.fields.values()] == [
        InputField,
        InputField,
    ]


def test_inputobjecttype_of_input():
    class Child(InputObjectType):
        first_name = String()
        last_name = String()

        @property
        def full_name(self):
            return f"{self.first_name} {self.last_name}"

    class Parent(InputObjectType):
        child = InputField(Child)

    class Query(ObjectType):
        is_child = Boolean(parent=Parent())

        def resolve_is_child(self, info, parent):
            return (
                isinstance(parent.child, Child)
                and parent.child.full_name == "Peter Griffin"
            )

    schema = Schema(query=Query)
    result = schema.execute(
        """query basequery {
        isChild(parent: {child: {firstName: "Peter", lastName: "Griffin"}})
    }
    """
    )

    assert not result.errors
    assert result.data == {"isChild": True}


def test_inputobjecttype_default_input_as_undefined(
    set_default_input_object_type_to_undefined,
):
    class TestUndefinedInput(InputObjectType):
        required_field = String(required=True)
        optional_field = String()

    class Query(ObjectType):
        undefined_optionals_work = Field(NonNull(Boolean), input=TestUndefinedInput())

        def resolve_undefined_optionals_work(self, info, input: TestUndefinedInput):
            # Confirm that optional_field comes as Undefined
            return (
                input.required_field == "required" and input.optional_field is Undefined
            )

    schema = Schema(query=Query)
    result = schema.execute(
        """query basequery {
        undefinedOptionalsWork(input: {requiredField: "required"})
    }
    """
    )

    assert not result.errors
    assert result.data == {"undefinedOptionalsWork": True}



from functools import partial

from pytest import raises

from ..scalars import String
from ..structures import List, NonNull
from .utils import MyLazyType


def test_list():
    _list = List(String)
    assert _list.of_type == String
    assert str(_list) == "[String]"


def test_list_with_unmounted_type():
    with raises(Exception) as exc_info:
        List(String())

    assert (
        str(exc_info.value)
        == "List could not have a mounted String() as inner type. Try with List(String)."
    )


def test_list_with_lazy_type():
    MyType = object()
    field = List(lambda: MyType)
    assert field.of_type == MyType


def test_list_with_lazy_partial_type():
    MyType = object()
    field = List(partial(lambda: MyType))
    assert field.of_type == MyType


def test_list_with_string_type():
    field = List("graphene.types.tests.utils.MyLazyType")
    assert field.of_type == MyLazyType


def test_list_inherited_works_list():
    _list = List(List(String))
    assert isinstance(_list.of_type, List)
    assert _list.of_type.of_type == String


def test_list_inherited_works_nonnull():
    _list = List(NonNull(String))
    assert isinstance(_list.of_type, NonNull)
    assert _list.of_type.of_type == String


def test_nonnull():
    nonnull = NonNull(String)
    assert nonnull.of_type == String
    assert str(nonnull) == "String!"


def test_nonnull_with_lazy_type():
    MyType = object()
    field = NonNull(lambda: MyType)
    assert field.of_type == MyType


def test_nonnull_with_lazy_partial_type():
    MyType = object()
    field = NonNull(partial(lambda: MyType))
    assert field.of_type == MyType


def test_nonnull_with_string_type():
    field = NonNull("graphene.types.tests.utils.MyLazyType")
    assert field.of_type == MyLazyType


def test_nonnull_inherited_works_list():
    _list = NonNull(List(String))
    assert isinstance(_list.of_type, List)
    assert _list.of_type.of_type == String


def test_nonnull_inherited_dont_work_nonnull():
    with raises(Exception) as exc_info:
        NonNull(NonNull(String))

    assert (
        str(exc_info.value)
        == "Can only create NonNull of a Nullable GraphQLType but got: String!."
    )


def test_nonnull_with_unmounted_type():
    with raises(Exception) as exc_info:
        NonNull(String())

    assert (
        str(exc_info.value)
        == "NonNull could not have a mounted String() as inner type. Try with NonNull(String)."
    )


def test_list_comparasion():
    list1 = List(String)
    list2 = List(String)
    list3 = List(None)

    list1_argskwargs = List(String, None, b=True)
    list2_argskwargs = List(String, None, b=True)

    assert list1 == list2
    assert list1 != list3
    assert list1_argskwargs == list2_argskwargs
    assert list1 != list1_argskwargs


def test_nonnull_comparasion():
    nonnull1 = NonNull(String)
    nonnull2 = NonNull(String)
    nonnull3 = NonNull(None)

    nonnull1_argskwargs = NonNull(String, None, b=True)
    nonnull2_argskwargs = NonNull(String, None, b=True)

    assert nonnull1 == nonnull2
    assert nonnull1 != nonnull3
    assert nonnull1_argskwargs == nonnull2_argskwargs
    assert nonnull1 != nonnull1_argskwargs



from functools import partial
from inspect import isclass

from ..types import Field, Interface, ObjectType
from ..types.interface import InterfaceOptions
from ..types.utils import get_type
from .id_type import BaseGlobalIDType, DefaultGlobalIDType


def is_node(objecttype):
    """
    Check if the given objecttype has Node as an interface
    """
    if not isclass(objecttype):
        return False

    if not issubclass(objecttype, ObjectType):
        return False

    return any(issubclass(i, Node) for i in objecttype._meta.interfaces)


class GlobalID(Field):
    def __init__(
        self,
        node=None,
        parent_type=None,
        required=True,
        global_id_type=DefaultGlobalIDType,
        *args,
        **kwargs,
    ):
        super(GlobalID, self).__init__(
            global_id_type.graphene_type, required=required, *args, **kwargs
        )
        self.node = node or Node
        self.parent_type_name = parent_type._meta.name if parent_type else None

    @staticmethod
    def id_resolver(parent_resolver, node, root, info, parent_type_name=None, **args):
        type_id = parent_resolver(root, info, **args)
        parent_type_name = parent_type_name or info.parent_type.name
        return node.to_global_id(parent_type_name, type_id)  # root._meta.name

    def wrap_resolve(self, parent_resolver):
        return partial(
            self.id_resolver,
            parent_resolver,
            self.node,
            parent_type_name=self.parent_type_name,
        )


class NodeField(Field):
    def __init__(self, node, type_=False, **kwargs):
        assert issubclass(node, Node), "NodeField can only operate in Nodes"
        self.node_type = node
        self.field_type = type_
        global_id_type = node._meta.global_id_type

        super(NodeField, self).__init__(
            # If we don't specify a type, the field type will be the node interface
            type_ or node,
            id=global_id_type.graphene_type(
                required=True, description="The ID of the object"
            ),
            **kwargs,
        )

    def wrap_resolve(self, parent_resolver):
        return partial(self.node_type.node_resolver, get_type(self.field_type))


class AbstractNode(Interface):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, global_id_type=DefaultGlobalIDType, **options):
        assert issubclass(
            global_id_type, BaseGlobalIDType
        ), "Custom ID type need to be implemented as a subclass of BaseGlobalIDType."
        _meta = InterfaceOptions(cls)
        _meta.global_id_type = global_id_type
        _meta.fields = {
            "id": GlobalID(
                cls, global_id_type=global_id_type, description="The ID of the object"
            )
        }
        super(AbstractNode, cls).__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    def resolve_global_id(cls, info, global_id):
        return cls._meta.global_id_type.resolve_global_id(info, global_id)


class Node(AbstractNode):
    """An object with an ID"""

    @classmethod
    def Field(cls, *args, **kwargs):  # noqa: N802
        return NodeField(cls, *args, **kwargs)

    @classmethod
    def node_resolver(cls, only_type, root, info, id):
        return cls.get_node_from_global_id(info, id, only_type=only_type)

    @classmethod
    def get_node_from_global_id(cls, info, global_id, only_type=None):
        _type, _id = cls.resolve_global_id(info, global_id)

        graphene_type = info.schema.get_type(_type)
        if graphene_type is None:
            raise Exception(f'Relay Node "{_type}" not found in schema')

        graphene_type = graphene_type.graphene_type

        if only_type:
            assert (
                graphene_type == only_type
            ), f"Must receive a {only_type._meta.name} id."

        # We make sure the ObjectType implements the "Node" interface
        if cls not in graphene_type._meta.interfaces:
            raise Exception(
                f'ObjectType "{_type}" does not implement the "{cls}" interface.'
            )

        get_node = getattr(graphene_type, "get_node", None)
        if get_node:
            return get_node(info, _id)

    @classmethod
    def to_global_id(cls, type_, id):
        return cls._meta.global_id_type.to_global_id(type_, id)



import inspect


def trim_docstring(docstring):
    # Cleans up whitespaces from an indented docstring
    #
    # See https://www.python.org/dev/peps/pep-0257/
    # and https://docs.python.org/2/library/inspect.html#inspect.cleandoc
    return inspect.cleandoc(docstring) if docstring else None
